# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

import os
import re
import time

import yaml

from .. import common, plugins
from ..asset import AppriseAsset
from ..manager_config import ConfigurationManager
from ..manager_plugins import NotificationManager
from ..url import URLBase
from ..utils.cwe312 import cwe312_url
from ..utils.parse import GET_SCHEMA_RE, parse_bool, parse_list, parse_urls
from ..utils.time import zoneinfo

# Test whether token is valid or not
VALID_TOKEN = re.compile(r"(?P<token>[a-z0-9][a-z0-9_]+)", re.I)

# Grant access to our Notification Manager Singleton
N_MGR = NotificationManager()

# Grant access to our Configuration Manager Singleton
C_MGR = ConfigurationManager()


class ConfigBase(URLBase):
    """This is the base class for all supported configuration sources."""

    # The Default Encoding to use if not otherwise detected
    encoding = "utf-8"

    # The default expected configuration format unless otherwise
    # detected by the sub-modules
    default_config_format = common.ConfigFormat.TEXT

    # This is only set if the user overrides the config format on the URL
    # this should always initialize itself as None
    config_format = None

    # Don't read any more of this amount of data into memory as there is no
    # reason we should be reading in more. This is more of a safe guard then
    # anything else. 128KB (131072B)
    max_buffer_size = 131072

    # By default all configuration is not includable using the 'include'
    # line found in configuration files.
    allow_cross_includes = common.ContentIncludeMode.NEVER

    # the config path manages the handling of relative include
    config_path = os.getcwd()

    def __init__(
        self,
        cache: bool | int = True,
        recursion: int = 0,
        insecure_includes: bool = False,
        **kwargs: object,
    ) -> None:
        """Initialize some general logging and common server arguments that
        will keep things consistent when working with the configurations that
        inherit this class.

        By default we cache our responses so that subsiquent calls does not
        cause the content to be retrieved again.  For local file references
        this makes no difference at all.  But for remote content, this does
        mean more then one call can be made to retrieve the (same) data.  This
        method can be somewhat inefficient if disabled.  Only disable caching
        if you understand the consequences.

        You can alternatively set the cache value to an int identifying the
        number of seconds the previously retrieved can exist for before it
        should be considered expired.

        recursion defines how deep we recursively handle entries that use the
        `include` keyword. This keyword requires us to fetch more configuration
        from another source and add it to our existing compilation. If the
        file we remotely retrieve also has an `include` reference, we will only
        advance through it if recursion is set to 2 deep.  If set to zero
        it is off.  There is no limit to how high you set this value. It would
        be recommended to keep it low if you do intend to use it.

        insecure_include by default are disabled. When set to True, all
        Apprise Config files marked to be in STRICT mode are treated as being
        in ALWAYS mode.

        Take a file:// based configuration for example, only a file:// based
        configuration can include another file:// based one. because it is set
        to STRICT mode. If an http:// based configuration file attempted to
        include a file:// one it woul fail. However this include would be
        possible if insecure_includes is set to True.

        There are cases where a self hosting apprise developer may wish to load
        configuration from memory (in a string format) that contains 'include'
        entries (even file:// based ones).  In these circumstances if you want
        these 'include' entries to be honored, this value must be set to True.
        """

        super().__init__(**kwargs)

        # Tracks the time the content was last retrieved on.  This place a role
        # for cases where we are not caching our response and are required to
        # re-retrieve our settings.
        self._cached_time = None

        # Tracks previously loaded content for speed
        self._cached_servers = None

        # Initialize our recursion value
        self.recursion = recursion

        # Initialize our insecure_includes flag
        self.insecure_includes = insecure_includes

        if "encoding" in kwargs:
            # Store the encoding
            self.encoding = kwargs.get("encoding")

        fmt = kwargs.get("format")
        if fmt:
            try:
                self.config_format = (
                    fmt if isinstance(fmt, common.ConfigFormat)
                    else common.ConfigFormat(fmt.lower())
                )

            except (AttributeError, ValueError):
                err = f"An invalid config format ({fmt}) was specified."
                self.logger.warning(err)
                raise TypeError(err) from None

        # Set our cache flag; it can be True or a (positive) integer
        try:
            self.cache = cache if isinstance(cache, bool) else int(cache)
            if self.cache < 0:
                err = f"A negative cache value ({cache}) was specified."
                self.logger.warning(err)
                raise TypeError(err)

        except (ValueError, TypeError):
            err = f"An invalid cache value ({cache}) was specified."
            self.logger.warning(err)
            raise TypeError(err) from None

        return

    def servers(
        self,
        asset: AppriseAsset | None = None,
        **kwargs: object,
    ) -> list[plugins.NotifyBase]:
        """Performs reads loaded configuration and returns all of the services
        that could be parsed and loaded."""

        if not self.expired():
            # We already have cached results to return; use them
            return self._cached_servers

        # Our cached response object
        self._cached_servers = []

        # read() causes the child class to do whatever it takes for the
        # config plugin to load the data source and return unparsed content
        # None is returned if there was an error or simply no data
        content = self.read(**kwargs)
        if not isinstance(content, str):
            # Set the time our content was cached at
            self._cached_time = time.time()

            # Nothing more to do; return our empty cache list
            return self._cached_servers

        # Our Configuration format uses a default if one wasn't one detected
        # or enfored.
        config_format = (
            self.default_config_format
            if self.config_format is None
            else self.config_format
        )

        # Dynamically load our parse_ function based on our config format
        fn = getattr(ConfigBase, f"config_parse_{config_format.value}")

        # Initialize our asset object
        asset = asset if isinstance(asset, AppriseAsset) else self.asset

        # Execute our config parse function which always returns a tuple
        # of our servers and our configuration
        servers, configs = fn(content=content, asset=asset)
        self._cached_servers.extend(servers)

        # Configuration files were detected; recursively populate them
        # If we have been configured to do so
        for url in configs:

            if self.recursion > 0:
                # Attempt to acquire the schema at the very least to allow
                # our configuration based urls.
                schema = GET_SCHEMA_RE.match(url)
                if schema is None:
                    # Plan B is to assume we're dealing with a file
                    schema = "file"
                    if not os.path.isabs(url):
                        # We're dealing with a relative path; prepend
                        # our current config path
                        url = os.path.join(self.config_path, url)

                    url = f"{schema}://{URLBase.quote(url)}"

                else:
                    # Ensure our schema is always in lower case
                    schema = schema.group("schema").lower()

                    # Some basic validation
                    if schema not in C_MGR:
                        ConfigBase.logger.warning(
                            f"Unsupported include schema {schema}."
                        )
                        continue

                # CWE-312 (Secure Logging) Handling
                loggable_url = (
                    url if not asset.secure_logging else cwe312_url(url)
                )

                # Parse our url details of the server object as dictionary
                # containing all of the information parsed from our URL
                results = C_MGR[schema].parse_url(url)
                if not results:
                    # Failed to parse the server URL
                    self.logger.warning(
                        f"Unparseable include URL {loggable_url}"
                    )
                    continue

                # Handle cross inclusion based on allow_cross_includes rules
                if (
                    C_MGR[schema].allow_cross_includes
                    == common.ContentIncludeMode.STRICT
                    and schema not in self.schemas()
                    and not self.insecure_includes
                ) or C_MGR[
                    schema
                ].allow_cross_includes == common.ContentIncludeMode.NEVER:

                    # Prevent the loading if insecure base protocols
                    ConfigBase.logger.warning(
                        f"Including {schema}:// based configuration is"
                        f" prohibited. Ignoring URL {loggable_url}"
                    )
                    continue

                # Prepare our Asset Object
                results["asset"] = asset

                # No cache is required because we're just lumping this in
                # and associating it with the cache value we've already
                # declared (prior to our recursion)
                results["cache"] = False

                # Recursion can never be parsed from the URL; we decrement
                # it one level
                results["recursion"] = self.recursion - 1

                # Insecure Includes flag can never be parsed from the URL
                results["insecure_includes"] = self.insecure_includes

                try:
                    # Attempt to create an instance of our plugin using the
                    # parsed URL information
                    cfg_plugin = C_MGR[results["schema"]](**results)

                except Exception as e:
                    # the arguments are invalid or can not be used.
                    self.logger.warning(
                        f"Could not load include URL: {loggable_url}"
                    )
                    self.logger.debug(f"Loading Exception: {e!s}")
                    continue

                # if we reach here, we can now add this servers found
                # in this configuration file to our list
                self._cached_servers.extend(cfg_plugin.servers(asset=asset))

                # We no longer need our configuration object
                del cfg_plugin

            else:
                # CWE-312 (Secure Logging) Handling
                loggable_url = (
                    url if not asset.secure_logging else cwe312_url(url)
                )

                self.logger.debug(
                    "Recursion limit reached; ignoring Include URL: %s",
                    loggable_url,
                )

        if self._cached_servers:
            self.logger.info(
                f"Loaded {len(self._cached_servers)} entries from"
                f" {self.url(privacy=asset.secure_logging)}"
            )
        else:
            self.logger.warning(
                "Failed to load Apprise configuration from"
                f" {self.url(privacy=asset.secure_logging)}"
            )

        # Set the time our content was cached at
        self._cached_time = time.time()

        return self._cached_servers

    def read(self) -> str | None:
        """This object should be implimented by the child classes."""
        return None

    def expired(self) -> bool:
        """Simply returns True if the configuration should be considered as
        expired or False if content should be retrieved."""
        if isinstance(self._cached_servers, list) and self.cache:
            # We have enough reason to look further into our cached content
            # and verify it has not expired.
            if self.cache is True:
                # we have not expired, return False
                return False

            # Verify our cache time to determine whether we will get our
            # content again.
            age_in_sec = time.time() - self._cached_time
            if age_in_sec <= self.cache:
                # We have not expired; return False
                return False

        # If we reach here our configuration should be considered
        # missing and/or expired.
        return True

    @staticmethod
    def __normalize_tag_groups(group_tags: dict[str, set[str]]) -> None:
        """
        Used to normalize a tag assign map which looks like:
          {
             'group': set('{tag1}', '{group1}', '{tag2}'),
             'group1': set('{tag2}','{tag3}'),
          }

          Then normalized it (merging groups); with respect to the above, the
          output would be:
          {
             'group': set('{tag1}', '{tag2}', '{tag3}),
             'group1': set('{tag2}','{tag3}'),
          }

        """
        # Prepare a key set list we can use
        tag_groups = {str(x) for x in group_tags}

        def _expand(tags, ignore=None):
            """Expands based on tag provided and returns a set.

            this also updates the group_tags while it goes
            """

            # Prepare ourselves a return set
            results = set()
            ignore = set() if ignore is None else ignore

            # track groups
            groups = set()

            for tag in tags:
                if tag in ignore:
                    continue

                # Track our groups
                groups.add(tag)

                # Store what we know is worth keeping
                if tag not in group_tags:  # pragma: no cover
                    # handle cases where the tag doesn't exist
                    group_tags[tag] = set()

                results |= group_tags[tag] - tag_groups

                # Get simple tag assignments
                found = group_tags[tag] & tag_groups
                if not found:
                    continue

                for gtag in found:
                    if gtag in ignore:
                        continue

                    # Go deeper (recursion)
                    ignore.add(tag)
                    group_tags[gtag] = _expand({gtag}, ignore=ignore)
                    results |= group_tags[gtag]

                    # Pop ignore
                    ignore.remove(tag)

            return results

        for tag in tag_groups:
            # Get our tags
            group_tags[tag] |= _expand({tag})
            if not group_tags[tag]:
                ConfigBase.logger.warning(
                    f"The group {tag} has no tags assigned to it"
                )
                del group_tags[tag]

    @staticmethod
    def parse_url(
        url: str,
        verify_host: bool = True,
    ) -> dict[str, object] | None:
        """Parses the URL and returns it broken apart into a dictionary.

        This is very specific and customized for Apprise.

        Args:
            url (str): The URL you want to fully parse.
            verify_host (:obj:`bool`, optional): a flag kept with the parsed
                 URL which some child classes will later use to verify SSL
                 keys (if SSL transactions take place).  Unless under very
                 specific circumstances, it is strongly recomended that
                 you leave this default value set to True.

        Returns:
            A dictionary is returned containing the URL fully parsed if
            successful, otherwise None is returned.
        """

        results = URLBase.parse_url(url, verify_host=verify_host)

        if not results:
            # We're done; we failed to parse our url
            return results

        # Allow overriding the default config format
        if "format" in results["qsd"]:
            results["format"] = results["qsd"].get("format")
            if results["format"] not in common.CONFIG_FORMATS:
                URLBase.logger.warning(
                    "Unsupported format specified {}".format(results["format"])
                )
                del results["format"]

        # Defines the encoding of the payload
        if "encoding" in results["qsd"]:
            results["encoding"] = results["qsd"].get("encoding")

        # Our cache value
        if "cache" in results["qsd"]:
            # First try to get it's integer value
            try:
                results["cache"] = int(results["qsd"]["cache"])

            except (ValueError, TypeError):
                # No problem, it just isn't an integer; now treat it as a bool
                # instead:
                results["cache"] = parse_bool(results["qsd"]["cache"])

        return results

    @staticmethod
    def detect_config_format(
        content: str,
        **kwargs: object,
    ) -> common.ConfigFormat | None:
        """Takes the specified content and attempts to detect the format type.

        The function returns the actual format type if detected, otherwise it
        returns None
        """

        # Detect Format Logic:
        #  - A pound/hashtag (#) is alawys a comment character so we skip over
        #     lines matched here.
        #  - Detection begins on the first non-comment and non blank line
        #     matched.
        #  - If we find a string followed by a colon, we know we're dealing
        #     with a YAML file.
        #  - If we find a string that starts with a URL, or our tag
        #     definitions (accepting commas) followed by an equal sign we know
        #     we're dealing with a TEXT format.

        # Define what a valid line should look like
        valid_line_re = re.compile(
            r"^\s*(?P<line>([;#]+(?P<comment>.*))|"
            r"(?P<text>((?P<tag>[ \t,a-z0-9_-]+)=)?[a-z0-9]+://.*)|"
            r"((?P<yaml>[a-z0-9]+):.*))?$",
            re.I,
        )

        try:
            # split our content up to read line by line
            content = re.split(r"\r*\n", content)

        except TypeError:
            # content was not expected string type
            ConfigBase.logger.error("Invalid Apprise configuration specified.")
            return None

        # By default set our return value to None since we don't know
        # what the format is yet
        config_format = None

        # iterate over each line of the file to attempt to detect it
        # stop the moment a the type has been determined
        for line, entry in enumerate(content, start=1):

            result = valid_line_re.match(entry)
            if not result:
                # Invalid syntax
                ConfigBase.logger.error(
                    "Undetectable Apprise configuration found "
                    f"based on line {line}."
                )
                # Take an early exit
                return None

            # Attempt to detect configuration
            if result.group("yaml"):
                config_format = common.ConfigFormat.YAML
                ConfigBase.logger.debug(
                    f"Detected YAML configuration based on line {line}."
                )
                break

            elif result.group("text"):
                config_format = common.ConfigFormat.TEXT
                ConfigBase.logger.debug(
                    f"Detected TEXT configuration based on line {line}."
                )
                break

            # If we reach here, we have a comment entry
            # Adjust default format to TEXT
            config_format = common.ConfigFormat.TEXT

        return config_format

    @staticmethod
    def config_parse(
        content: str,
        asset: AppriseAsset | None = None,
        config_format: str | common.ConfigFormat | None = None,
        **kwargs: object,
    ) -> tuple[list[object], list[str]]:
        """Takes the specified config content and loads it based on the
        specified config_format.

        If a format isn't specified, then it is auto detected.
        """

        if config_format is None:
            # Detect the format
            config_format = ConfigBase.detect_config_format(content)

            if not config_format:
                # We couldn't detect configuration
                ConfigBase.logger.error("Could not detect configuration")
                return ([], [])

        try:
            config_format = (
                config_format if isinstance(config_format, common.ConfigFormat)
                else common.ConfigFormat(config_format.lower())
            )

        except (AttributeError, ValueError):
            ConfigBase.logger.error(
                f"An invalid configuration format ({config_format}) was"
                " specified"
            )
            return ([], [])

        # Dynamically load our parse_ function based on our config format
        fn = getattr(ConfigBase, f"config_parse_{config_format.value}")

        # Execute our config parse function which always returns a list
        return fn(content=content, asset=asset)

    @staticmethod
    def config_parse_text(
        content: str,
        asset: AppriseAsset | None = None,
    ) -> tuple[list[object], list[str]]:
        """Parse the specified content as though it were a simple text file
        only containing a list of URLs.

        Return a tuple that looks like (servers, configs) where:
          - servers contains a list of loaded notification plugins
          - configs contains a list of additional configuration files
            referenced.

        You may also optionally associate an asset with the notification.

        The file syntax is:

            #
            # pound/hashtag allow for line comments
            #
            # One or more tags can be idenified using comma's (,) to separate
            # them.
            <Tag(s)>=<URL>

            # Or you can use this format (no tags associated)
            <URL>

            # you can also use the keyword 'include' and identify a
            # configuration location (like this file) which will be included
            # as additional configuration entries when loaded.
            include <ConfigURL>

            # Assign tag contents to a group identifier
            <Group(s)>=<Tag(s)>
        """
        # A list of loaded Notification Services
        servers = []

        # A list of additional configuration files referenced using
        # the include keyword
        configs = []

        # Track all of the tags we want to assign later on
        group_tags = {}

        # Track our entries to preload
        preloaded = []

        # Prepare our Asset Object
        asset = asset if isinstance(asset, AppriseAsset) else AppriseAsset()

        # Define what a valid line should look like
        valid_line_re = re.compile(
            r"^\s*(?P<line>([;#]+(?P<comment>.*))|"
            r"(\s*(?P<tags>[a-z0-9, \t_-]+)\s*=|=)?\s*"
            r"((?P<url>[a-z0-9]{1,32}://.*)|(?P<assign>[a-z0-9, \t_-]+))|"
            r"include\s+(?P<config>.+))?\s*$",
            re.I,
        )

        try:
            # split our content up to read line by line
            content = re.split(r"\r*\n", content)

        except TypeError:
            # content was not expected string type
            ConfigBase.logger.error(
                "Invalid Apprise TEXT based configuration specified."
            )
            return ([], [])

        for line, entry in enumerate(content, start=1):
            result = valid_line_re.match(entry)
            if not result:
                # Invalid syntax
                ConfigBase.logger.error(
                    "Invalid Apprise TEXT configuration format found "
                    f"{entry} on line {line}."
                )

                # Assume this is a file we shouldn't be parsing. It's owner
                # can read the error printed to screen and take action
                # otherwise.
                return ([], [])

            # Retrieve our line
            url, assign, config = (
                result.group("url"),
                result.group("assign"),
                result.group("config"),
            )

            if not (url or config or assign):
                # Comment/empty line; do nothing
                continue

            if config:
                # CWE-312 (Secure Logging) Handling
                loggable_url = (
                    config if not asset.secure_logging else cwe312_url(config)
                )

                ConfigBase.logger.debug(f"Include URL: {loggable_url}")

                # Store our include line
                configs.append(config.strip())
                continue

            # CWE-312 (Secure Logging) Handling
            loggable_url = url if not asset.secure_logging else cwe312_url(url)

            if assign:
                groups = set(parse_list(result.group("tags"), cast=str))
                if not groups:
                    # no tags were assigned
                    ConfigBase.logger.warning(
                        "Unparseable tag assignment - no group(s) "
                        f"on line {line}"
                    )
                    continue

                # Get our tags
                tags = set(parse_list(assign, cast=str))
                if not tags:
                    # no tags were assigned
                    ConfigBase.logger.warning(
                        "Unparseable tag assignment - no tag(s) to assign "
                        f"on line {line}"
                    )
                    continue

                # Update our tag group map
                for tag_group in groups:
                    if tag_group not in group_tags:
                        group_tags[tag_group] = set()

                    # ensure our tag group is never included in the assignment
                    group_tags[tag_group] |= tags - {tag_group}
                continue

            # Acquire our url tokens
            results = plugins.url_to_dict(
                url, secure_logging=asset.secure_logging
            )
            if results is None:
                # Failed to parse the server URL
                ConfigBase.logger.warning(
                    f"Unparseable URL {loggable_url} on line {line}."
                )
                continue

            # Build a list of tags to associate with the newly added
            # notifications if any were set
            results["tag"] = set(parse_list(result.group("tags"), cast=str))

            # Set our Asset Object
            results["asset"] = asset

            # Store our preloaded entries
            preloaded.append({
                "results": results,
                "line": line,
                "loggable_url": loggable_url,
            })

        #
        # Normalize Tag Groups
        # - Expand Groups of Groups so that they don't exist
        #
        ConfigBase.__normalize_tag_groups(group_tags)

        #
        # URL Processing
        #
        for entry in preloaded:
            # Point to our results entry for easier reference below
            results = entry["results"]

            #
            # Apply our tag groups if they're defined
            #
            for group, tags in group_tags.items():
                # Detect if anything assigned to this tag also maps back to a
                # group.  If so we want to add the group to our list
                if next(
                    (True for tag in results["tag"] if tag in tags), False
                ):
                    results["tag"].add(group)

            try:
                # Attempt to create an instance of our plugin using the
                # parsed URL information
                plugin = N_MGR[results["schema"]](**results)

                # Create log entry of loaded URL
                ConfigBase.logger.debug(
                    "Loaded URL: %s",
                    plugin.url(privacy=results["asset"].secure_logging),
                )

            except Exception as e:
                # the arguments are invalid or can not be used.
                ConfigBase.logger.warning(
                    "Could not load URL {} on line {}.".format(
                        entry["loggable_url"], entry["line"]
                    )
                )
                ConfigBase.logger.debug(f"Loading Exception: {e!s}")
                continue

            # if we reach here, we successfully loaded our data
            servers.append(plugin)

        # Return what was loaded
        return (servers, configs)

    @staticmethod
    def config_parse_yaml(
        content: str,
        asset: AppriseAsset | None = None,
    ) -> tuple[list[object], list[str]]:
        """Parse the specified content as though it were a yaml file
        specifically formatted for Apprise.

        Return a tuple that looks like (servers, configs) where:
          - servers contains a list of loaded notification plugins
          - configs contains a list of additional configuration files
            referenced.

        You may optionally associate an asset with the notification.
        """

        # A list of loaded Notification Services
        servers = []

        # A list of additional configuration files referenced using
        # the include keyword
        configs = []

        # Group Assignments
        group_tags = {}

        # Track our entries to preload
        preloaded = []

        try:
            # Load our data (safely)
            result = yaml.load(content, Loader=yaml.SafeLoader)

        except (
            AttributeError,
            yaml.parser.ParserError,
            yaml.error.MarkedYAMLError,
        ) as e:
            # Invalid content
            ConfigBase.logger.error("Invalid Apprise YAML data specified.")
            ConfigBase.logger.debug(f"YAML Exception:{os.linesep}{e}")
            return ([], [])

        if not isinstance(result, dict):
            # Invalid content
            ConfigBase.logger.error(
                "Invalid Apprise YAML based configuration specified."
            )
            return ([], [])

        # YAML Version
        version = result.get("version", 1)
        if version != 1:
            # Invalid syntax
            ConfigBase.logger.error(
                f"Invalid Apprise YAML version specified {version}."
            )
            return ([], [])

        #
        # global asset object
        #
        asset = asset if isinstance(asset, AppriseAsset) else AppriseAsset()

        # Prepare our default timezone
        default_timezone = asset.tzinfo

        # Acquire our asset tokens
        tokens = result.get("asset", None)
        if tokens and isinstance(tokens, dict):
            raw_tz = tokens.get("timezone", tokens.get("tz"))
            if isinstance(raw_tz, str):
                default_timezone = zoneinfo(re.sub(r"[^\w/-]+", "", raw_tz))
                if not default_timezone:
                    ConfigBase.logger.warning(
                        'Ignored invalid timezone "%s"', raw_tz)
                    default_timezone = asset.tzinfo
                else:
                    asset._tzinfo = default_timezone

            elif raw_tz is not None:
                ConfigBase.logger.warning(
                    'Ignored invalid timezone "%r"', raw_tz)

            # Iterate over remaining tokens
            for k, v in tokens.items():
                if k.startswith("_") or k.endswith("_"):
                    # Entries are considered reserved if they start or end
                    # with an underscore
                    ConfigBase.logger.warning(f'Ignored asset key "{k}".')
                    continue

                if not (
                    hasattr(asset, k)
                    and isinstance(getattr(asset, k), (bool, str))
                ):

                    # We can't set a function or non-string set value
                    ConfigBase.logger.warning(f'Invalid asset key "{k}".')
                    continue

                if v is None:
                    # Convert to an empty string
                    v = ""

                if isinstance(v, (bool, str)) and isinstance(
                    getattr(asset, k), bool
                ):

                    # If the object in the Asset is a boolean, then
                    # we want to convert the specified string to
                    # match that.
                    setattr(asset, k, parse_bool(v))

                elif isinstance(v, str):
                    # Set our asset object with the new value
                    setattr(asset, k, v.strip())

                else:
                    # we must set strings with a string
                    ConfigBase.logger.warning(f'Invalid asset value to "{k}".')
                    continue
        #
        # global tag root directive
        #
        global_tags = set()

        tags = result.get("tag", None)
        if tags and isinstance(tags, (list, tuple, str)):
            # Store any preset tags
            global_tags = set(parse_list(tags, cast=str))

        #
        # groups root directive
        #
        groups = result.get("groups", None)
        if isinstance(groups, dict):
            #
            # Dictionary
            #
            for _groups, tags in groups.items():
                for group in parse_list(_groups, cast=str):
                    if isinstance(tags, (list, tuple)):
                        _tags = set()
                        for e in tags:
                            if isinstance(e, dict):
                                _tags |= set(e.keys())
                            else:
                                _tags |= set(parse_list(e, cast=str))

                        # Final assignment
                        tags = _tags

                    else:
                        tags = set(parse_list(tags, cast=str))

                    if group not in group_tags:
                        group_tags[group] = tags

                    else:
                        group_tags[group] |= tags

        elif isinstance(groups, (list, tuple)):
            #
            # List of Dictionaries
            #

            # Iterate over each group defined and store it
            for no, entry in enumerate(groups):
                if not isinstance(entry, dict):
                    ConfigBase.logger.warning(
                        f"No assignment for group {entry}, entry #{no + 1}"
                    )
                    continue

                for _groups, tags in entry.items():
                    for group in parse_list(_groups, cast=str):
                        if isinstance(tags, (list, tuple)):
                            _tags = set()
                            for e in tags:
                                if isinstance(e, dict):
                                    _tags |= set(e.keys())
                                else:
                                    _tags |= set(parse_list(e, cast=str))

                            # Final assignment
                            tags = _tags

                        else:
                            tags = set(parse_list(tags, cast=str))

                        if group not in group_tags:
                            group_tags[group] = tags

                        else:
                            group_tags[group] |= tags

        # include root directive
        #
        includes = result.get("include", None)
        if isinstance(includes, str):
            # Support a single inline string or multiple ones separated by a
            # comma and/or space
            includes = parse_urls(includes)

        elif not isinstance(includes, (list, tuple)):
            # Not a problem; we simply have no includes
            includes = []

        # Iterate over each config URL
        for _no, url in enumerate(includes):

            if isinstance(url, str):
                # Support a single inline string or multiple ones separated by
                # a comma and/or space
                configs.extend(parse_urls(url))

            elif isinstance(url, dict):
                # Store the url and ignore arguments associated
                configs.extend(u for u in url)

        #
        # urls root directive
        #
        urls = result.get("urls", None)
        if not isinstance(urls, (list, tuple)):
            # Not a problem; we simply have no urls
            urls = []

        # Iterate over each URL
        for no, url in enumerate(urls):

            # Our results object is what we use to instantiate our object if
            # we can. Reset it to None on each iteration
            results = []

            # CWE-312 (Secure Logging) Handling
            loggable_url = url if not asset.secure_logging else cwe312_url(url)

            if isinstance(url, str):
                # We're just a simple URL string...
                schema = GET_SCHEMA_RE.match(url)
                if schema is None:
                    # Log invalid entries so that maintainer of config
                    # config file at least has something to take action
                    # with.
                    ConfigBase.logger.warning(
                        f"Invalid URL {loggable_url}, entry #{no + 1}"
                    )
                    continue

                # We found a valid schema worthy of tracking; store it's
                # details:
                _results = plugins.url_to_dict(
                    url, secure_logging=asset.secure_logging
                )
                if _results is None:
                    ConfigBase.logger.warning(
                        f"Unparseable URL {loggable_url}, entry #{no + 1}"
                    )
                    continue

                # add our results to our global set
                results.append(_results)

            elif isinstance(url, dict):
                # We are a url string with additional unescaped options. In
                # this case we want to iterate over all of our options so we
                # can at least tell the end user what entries were ignored
                # due to errors

                it = iter(url.items())

                # Track the URL to-load
                _url = None

                # Track last acquired schema
                schema = None

                for key, _tokens in it:
                    # Test our schema
                    _schema = GET_SCHEMA_RE.match(key)
                    if _schema is None:
                        # Log invalid entries so that maintainer of config
                        # config file at least has something to take action
                        # with.
                        ConfigBase.logger.warning(
                            f"Ignored entry {key} found under urls, entry"
                            f" #{no + 1}"
                        )
                        continue

                    # Store our schema
                    schema = _schema.group("schema").lower()

                    # Store our URL and Schema Regex
                    _url = key

                    # Update our token assignment
                    tokens = _tokens

                    # We're done
                    break

                if _url is None:
                    # the loop above failed to match anything
                    ConfigBase.logger.warning(
                        f"Unsupported URL, entry #{no + 1}"
                    )
                    continue

                _results = plugins.url_to_dict(
                    _url, secure_logging=asset.secure_logging
                )
                if _results is None:
                    # Setup dictionary
                    _results = {
                        # Minimum requirements
                        "schema": schema,
                    }

                if isinstance(tokens, (list, tuple, set)):
                    # populate and/or override any results populated by
                    # parse_url()
                    for entries in tokens:
                        # Copy ourselves a template of our parsed URL as a base
                        # to work with
                        r = _results.copy()

                        # We are a url string with additional unescaped options
                        if isinstance(entries, dict):
                            _url, tokens = next(iter(url.items()))

                            # Tags you just can't over-ride
                            if "schema" in entries:
                                del entries["schema"]

                            # support our special tokens (if they're present)
                            if schema in N_MGR:
                                entries = ConfigBase._special_token_handler(
                                    schema, entries
                                )

                            # Extend our dictionary with our new entries
                            r.update(entries)

                            # add our results to our global set
                            results.append(r)

                elif isinstance(tokens, dict):
                    # support our special tokens (if they're present)
                    if schema in N_MGR:
                        tokens = ConfigBase._special_token_handler(
                            schema, tokens
                        )

                    # Copy ourselves a template of our parsed URL as a base to
                    # work with
                    r = _results.copy()

                    # add our result set
                    r.update(tokens)

                    # add our results to our global set
                    results.append(r)

                else:
                    # add our results to our global set
                    results.append(_results)

            else:
                # Unsupported
                ConfigBase.logger.warning(
                    f"Unsupported Apprise YAML entry #{no + 1}"
                )
                continue

            # Track our entries
            entry = 0

            while len(results):
                # Increment our entry count
                entry += 1

                # Grab our first item
                _results = results.pop(0)

                if _results["schema"] not in N_MGR:
                    # the arguments are invalid or can not be used.
                    ConfigBase.logger.warning(
                        "An invalid Apprise schema ({}) in YAML configuration "
                        "entry #{}, item #{}".format(
                            _results["schema"], no + 1, entry
                        )
                    )
                    continue

                # tag is a special keyword that is managed by Apprise object.
                # The below ensures our tags are set correctly
                if "tag" in _results:
                    # Tidy our list up
                    _results["tag"] = (
                        set(parse_list(_results["tag"], cast=str))
                        | global_tags
                    )

                else:
                    # Just use the global settings
                    _results["tag"] = global_tags

                for key in list(_results.keys()):
                    # Strip out any tokens we know that we can't accept and
                    # warn the user
                    match = VALID_TOKEN.match(key)
                    if not match:
                        ConfigBase.logger.warning(
                            f"Ignoring invalid token ({key}) found in YAML "
                            f"configuration entry #{no + 1}, item #{entry}"
                        )
                        del _results[key]

                ConfigBase.logger.trace(
                    "URL #{}: {} unpacked as:{}{}".format(
                        no + 1,
                        url,
                        os.linesep,
                        os.linesep.join(
                            [f'{k}="{a}"' for k, a in _results.items()]
                        ),
                    )
                )

                # Prepare our Asset Object
                _results["asset"] = asset

                # Handle post processing of result set
                _results = URLBase.post_process_parse_url_results(_results)

                # Store our preloaded entries
                preloaded.append({
                    "results": _results,
                    "entry": no + 1,
                    "item": entry,
                })

        #
        # Normalize Tag Groups
        # - Expand Groups of Groups so that they don't exist
        #
        ConfigBase.__normalize_tag_groups(group_tags)

        #
        # URL Processing
        #
        for entry in preloaded:
            # Point to our results entry for easier reference below
            results = entry["results"]

            #
            # Apply our tag groups if they're defined
            #
            for group, tags in group_tags.items():
                # Detect if anything assigned to this tag also maps back to a
                # group.  If so we want to add the group to our list
                if next(
                    (True for tag in results["tag"] if tag in tags), False
                ):
                    results["tag"].add(group)

            # Now we generate our plugin
            try:
                # Attempt to create an instance of our plugin using the
                # parsed URL information
                plugin = N_MGR[results["schema"]](**results)

                # Create log entry of loaded URL
                ConfigBase.logger.debug(
                    "Loaded URL: %s",
                    plugin.url(privacy=results["asset"].secure_logging),
                )

            except Exception as e:
                # the arguments are invalid or can not be used.
                ConfigBase.logger.warning(
                    "Could not load Apprise YAML configuration "
                    "entry #{}, item #{}".format(entry["entry"], entry["item"])
                )
                ConfigBase.logger.debug(f"Loading Exception: {e!s}")
                continue

            # if we reach here, we successfully loaded our data
            servers.append(plugin)

        return (servers, configs)

    def pop(self, index: int = -1) -> object:
        """Removes an indexed Notification Service from the stack and returns
        it.

        By default, the last element of the list is removed.
        """

        if not isinstance(self._cached_servers, list):
            # Generate ourselves a list of content we can pull from
            self.servers()

        # Pop the element off of the stack
        return self._cached_servers.pop(index)

    @staticmethod
    def _special_token_handler(
        schema: str,
        tokens: dict[str, object],
    ) -> dict[str, object]:
        """This function takes a list of tokens and updates them to no longer
        include any special tokens such as +,-, and :

        - schema must be a valid schema of a supported plugin type
        - tokens must be a dictionary containing the yaml entries parsed.

        The idea here is we can post process a set of tokens provided in
        a YAML file where the user provided some of the special keywords.

        We effectivley look up what these keywords map to their appropriate
        value they're expected
        """
        # Create a copy of our dictionary
        tokens = tokens.copy()

        for kw, meta in N_MGR[schema].template_kwargs.items():

            # Determine our prefix:
            prefix = meta.get("prefix", "+")

            # Detect any matches
            matches = {
                k[1:]: str(v)
                for k, v in tokens.items()
                if k.startswith(prefix)
            }

            if not matches:
                # we're done with this entry
                continue

            if not isinstance(tokens.get(kw), dict):
                # Invalid; correct it
                tokens[kw] = {}

            # strip out processed tokens
            tokens = {
                k: v for k, v in tokens.items() if not k.startswith(prefix)
            }

            # Update our entries
            tokens[kw].update(matches)

        # Now map our tokens accordingly to the class templates defined by
        # each service.
        #
        # This is specifically used for YAML file parsing.  It allows a user to
        # define an entry such as:
        #
        # urls:
        #   - mailto://user:pass@domain:
        #       - to: user1@hotmail.com
        #       - to: user2@hotmail.com
        #
        # Under the hood, the NotifyEmail() class does not parse the `to`
        # argument. It's contents needs to be mapped to `targets`.  This is
        # defined in the class via the `template_args` and template_tokens`
        # section.
        #
        # This function here allows these mappings to take place within the
        # YAML file as independant arguments.
        class_templates = plugins.details(N_MGR[schema])

        for key in list(tokens.keys()):

            if key not in class_templates["args"]:
                # No need to handle non-arg entries
                continue

            # get our `map_to` and/or 'alias_of' value (if it exists)
            map_to = class_templates["args"][key].get(
                "alias_of", class_templates["args"][key].get("map_to", "")
            )

            if map_to == key:
                # We're already good as we are now
                continue

            if map_to in class_templates["tokens"]:
                meta = class_templates["tokens"][map_to]

            else:
                meta = class_templates["args"].get(
                    map_to, class_templates["args"][key]
                )

            # Perform a translation/mapping if our code reaches here
            value = tokens[key]
            del tokens[key]

            # Detect if we're dealign with a list or not
            is_list = re.search(r"^list:.*", meta.get("type"), re.IGNORECASE)

            if map_to not in tokens:
                tokens[map_to] = [] if is_list else meta.get("default")

            elif is_list and not isinstance(tokens.get(map_to), list):
                # Convert ourselves to a list if we aren't already
                tokens[map_to] = [tokens[map_to]]

            # Type Conversion
            if re.search(
                r"^(choice:)?string", meta.get("type"), re.IGNORECASE
            ) and not isinstance(value, str):

                # Ensure our format is as expected
                value = str(value)

            # Apply any further translations if required (absolute map)
            # This is the case when an arg maps to a token which further
            # maps to a different function arg on the class constructor
            abs_map = meta.get("map_to", map_to)

            # Set our token as how it was provided by the configuration
            if isinstance(tokens.get(map_to), list):
                tokens[abs_map].append(value)

            else:
                tokens[abs_map] = value

        # Return our tokens
        return tokens

    def __getitem__(self, index: int) -> object:
        """Returns the indexed server entry associated with the loaded
        notification servers."""
        if not isinstance(self._cached_servers, list):
            # Generate ourselves a list of content we can pull from
            self.servers()

        return self._cached_servers[index]

    def __iter__(self) -> object:
        """Returns an iterator to our server list."""
        if not isinstance(self._cached_servers, list):
            # Generate ourselves a list of content we can pull from
            self.servers()

        return iter(self._cached_servers)

    def __len__(self) -> int:
        """Returns the total number of servers loaded."""
        if not isinstance(self._cached_servers, list):
            # Generate ourselves a list of content we can pull from
            self.servers()

        return len(self._cached_servers)

    def __bool__(self) -> bool:
        """Allows the Apprise object to be wrapped in an 'if statement'.

        True is returned if our content was downloaded correctly.
        """
        if not isinstance(self._cached_servers, list):
            # Generate ourselves a list of content we can pull from
            self.servers()

        return bool(self._cached_servers)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from itertools import chain
from json import loads
import re

import requests

from ..common import NotifyType, PersistentStoreMode
from ..exception import AppriseException
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_list
from .base import NotifyBase

# Is Group Detection
IS_GROUP = re.compile(
    r"^\s*((#|%23)(?P<group>[a-z0-9_-]+)|"
    r"((#|%23)?(?P<all>all|everyone|\*)))\s*$",
    re.I,
)

# Is User Detection
IS_USER = re.compile(
    r"^\s*(@|%40)?(?P<user>[a-z0-9_-]+)\s*$",
    re.I,
)


class NextcloudGroupDiscoveryException(AppriseException):
    """Apprise Nextcloud Group Discovery Exception Class."""


class NotifyNextcloud(NotifyBase):
    """A wrapper for Nextcloud Notifications.

    Targets can be individual users, groups, or everyone:
    - user: specify one or more usernames as path segments
    - group: prefix with a hash (e.g., ``#DevTeam``)
    - everyone: use ``all`` (aliases: ``everyone``, ``*``)

    Group and everyone expansion uses Nextcloud's OCS provisioning API and
    requires appropriate permissions (typically an admin account) and the
    provisioning API enabled on the server.
    """

    # The default descriptive name associated with the Notification
    service_name = "Nextcloud"

    # The services URL
    service_url = "https://nextcloud.com/"

    # Insecure protocol (for those self hosted requests)
    protocol = "ncloud"

    # The default protocol (this is secure for notica)
    secure_protocol = "nclouds"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_nextcloud"

    # Nextcloud title length
    title_maxlen = 255

    # Defines the maximum allowable characters per message.
    body_maxlen = 4000

    # Our default is to not use persistent storage beyond in-memory
    # reference
    storage_mode = PersistentStoreMode.AUTO

    # Defines how long we cache our discovery for
    group_discovery_cache_length_sec = 86400

    # unique identifier to cache the 'all' group category
    all_group_id = "all"

    # Define object templates
    templates = (
        "{schema}://{host}/{targets}",
        "{schema}://{host}:{port}/{targets}",
        "{schema}://{user}:{password}@{host}/{targets}",
        "{schema}://{user}:{password}@{host}:{port}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "map_to": "targets",
                "prefix": "@",
            },
            "target_group": {
                "name": _("Target Group"),
                "type": "string",
                "map_to": "targets",
                "prefix": "#",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            # Nextcloud uses different API end points depending on the version
            # being used however the (API) payload remains the same.  Allow
            # users to specify the version they are using:
            "version": {
                "name": _("Version"),
                "type": "int",
                "min": 1,
                "default": 21,
            },
            "url_prefix": {
                "name": _("URL Prefix"),
                "type": "string",
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
    }

    def __init__(
        self,
        targets=None,
        version=None,
        headers=None,
        url_prefix=None,
        **kwargs,
    ):
        """Initialize Nextcloud Object."""
        super().__init__(**kwargs)

        # Store our targets
        self.targets = []
        self.groups = set()

        for target in parse_list(targets):
            results = IS_GROUP.match(target)
            if results:
                group_id = (
                    self.all_group_id
                    if results.group("all") else results.group("group"))

                self.groups.add(group_id)
                self.logger.debug("Added Nextcloud group '%s'", group_id)
                continue

            results = IS_USER.match(target)
            if results:
                # Store our target
                self.targets.append(results.group("user"))
                self.logger.debug(
                    "Added Nextcloud user '%s'", self.targets[-1])
                continue

            self.logger.warning(
                "Ignored invalid Nextcloud user/group '%s'", target)

        self.version = self.template_args["version"]["default"]
        if version is not None:
            try:
                self.version = int(version)
                if self.version < self.template_args["version"]["min"]:
                    # Let upper exception handle this
                    raise ValueError()

            except (ValueError, TypeError):
                msg = (
                    f"At invalid Nextcloud version ({version}) was specified."
                )
                self.logger.warning(msg)
                raise TypeError(msg) from None

        # Support URL Prefix
        self.url_prefix = "" if not url_prefix else (
            "/" + url_prefix.strip("/"))

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        return

    def _fetch(self, payload=None, target=None, group=None):
        """Wrapper to NextCloud API requests object."""

        # our method
        method = "POST" if target else "GET"

        # Prepare our Header
        headers = {
            "User-Agent": self.app_id,
            "OCS-APIREQUEST": "true",
            "Accept": "application/json",
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        # Prepare base URL fragments
        scheme = "https" if self.secure else "http"
        host_port = (
            self.host
            if not isinstance(self.port, int)
            else f"{self.host}:{self.port}"
        )
        base = f"{scheme}://{host_port}"
        if self.url_prefix:
            base = f"{base}{self.url_prefix}"

        # Auth
        auth = (self.user, self.password) if self.user else None

        # our URL Parameters
        params = {}

        # our response
        content = None

        if target:
            # Nextcloud URL based on version used
            query = f'v{self.version} Notify "{target}"'
            esc_target = NotifyNextcloud.quote(target)
            url = (
                f"{base}/ocs/v2.php/"
                "apps/admin_notifications/"
                f"api/v1/notifications/{esc_target}"
                if self.version < 21
                else (
                    f"{base}/ocs/v2.php/"
                    "apps/notifications/"
                    f"api/v2/admin_notifications/{esc_target}"
                )
            )

        elif group:
            query = f'Group "{group}"'
            params = {
                "format": "json",
            }
            esc_group = NotifyNextcloud.quote(group)
            url = f"{base}/ocs/v1.php/cloud/groups/{esc_group}"

        else:  # Users
            query = "Users"
            params = {
                "format": "json",
            }
            url = f"{base}/ocs/v1.php/cloud/users"

        self.throttle()
        self.logger.debug(
            "Nextcloud %s %s URL: %s (cert_verify=%r)",
            query,
            method,
            url,
            self.verify_certificate,
        )

        if payload:
            self.logger.debug(
                "Nextcloud v%d Payload: %s", self.version, str(payload)
            )

        try:
            # Prepare our request object
            request = requests.post if target else requests.get

            r = request(
                url,
                headers=headers,
                data=payload,
                params=params,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                content = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                content = {}

            if r.status_code != requests.codes.ok:
                status_str = NotifyNextcloud.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Nextcloud %s: %s%serror=%d.",
                    query,
                    status_str,
                    ", " if status_str else "",
                    r.status_code,
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")
                if target:
                    return (False, content)

                raise NextcloudGroupDiscoveryException(
                    f"{query} non-200 response")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred with Nextcloud %s",
                query,
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            if target:
                return (False, None)
            raise NextcloudGroupDiscoveryException(
                f"{query} socket exception") from None

        self.logger.info("Sent Nextcloud %s", query)
        return (True, content)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Nextcloud Notification."""

        # Create a copy of our targets
        targets = set(self.targets)

        # Initialize our has_error flag
        has_error = False

        if self.groups:
            # Append our group lookup
            try:
                for group in self.groups:
                    if group == self.all_group_id:
                        targets |= self.all_users()

                    else:
                        # specific group
                        targets |= self.users_by_group(group)

            except NextcloudGroupDiscoveryException:
                # logging already handled within all_users and user_by_group()
                return False

        if not targets:
            # There were no services to notify
            self.logger.warning("There were no Nextcloud targets to notify.")
            return False

        for target in targets:
            # Prepare our Payload
            payload = {
                "shortMessage": title if title else self.app_desc,
            }
            if body:
                # Only store the longMessage if a body was defined; nextcloud
                # doesn't take kindly to empty longMessage entries.
                payload["longMessage"] = body

            is_okay, _ = self._fetch(payload, target)
            if not is_okay:
                # Toggle our status
                has_error = True

        return not has_error

    def users_by_group(self, group):
        """
        Lists users associated with a provided group
        """

        # Check our cache
        targets = self.store.get(group)
        if targets is not None:
            # Returned cached value
            self.logger.trace(
                f"Using Nextcloud cached response for group '{group}' "
                "query")
            return set(targets)

        # _fetch throws an exception if it fails, so we can
        # go ahead and ignore checking for it.
        _, response = self._fetch(group=group)

        # Initialize our targets
        targets = set()

        # If we get here, our fetch was successful; look up our users
        users = response.get("ocs", {}).get("data", {}).get("users")
        if isinstance(users, list):
            targets = {
                s for u in users if (s := str(u).strip())
            }

        if not targets:
            self.logger.warning(
                "No users associated with Nextcloud group '%s'", group
            )

        self.store.set(
            group, list(targets),
            expires=self.group_discovery_cache_length_sec)
        return targets

    def all_users(self):
        """
        Lists users associated with Nextcloud instance
        """
        # Check our cache
        targets = self.store.get(self.all_group_id)
        if targets is not None:
            self.logger.trace(
                "Using Nextcloud cached response for all-user query")
            return set(targets)

        # _fetch throws an exception if it fails, so we can
        # go ahead and ignore checking for it.
        _, response = self._fetch()

        # Initialize our targets
        targets = set()

        # If we get here, our fetch was successful; look up our users
        users = response.get("ocs", {}).get("data", {}).get("users")
        if isinstance(users, list):
            targets = {
                s for u in users if (s := str(u).strip())
            }

        if not targets:
            self.logger.warning(
                "Failed to retrieve all users from Nextcloud",
            )
            # early exit; no cache
            return targets

        self.store.set(
            self.all_group_id,
            list(targets), expires=self.group_discovery_cache_length_sec)
        return targets

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
            self.url_prefix,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Create URL parameters from our headers
        params = {f"+{k}": v for k, v in self.headers.items()}

        # Set our version
        params["version"] = str(self.version)

        if self.url_prefix.rstrip("/"):
            params["url_prefix"] = self.url_prefix

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyNextcloud.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyNextcloud.quote(self.user, safe=""),
            )

        group_prefix = self.template_tokens["target_group"]["prefix"]
        user_prefix = self.template_tokens["target_user"]["prefix"]

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}/{targets}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a
            # valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            targets="/".join([
                NotifyNextcloud.quote(x, safe=(group_prefix + user_prefix))
                for x in chain(
                    # Groups are prefixed with a pound/hashtag symbol
                    [f"{group_prefix}{x}" for x in self.groups],
                    # Users
                    [f"{user_prefix}{x}" for x in self.targets],
                )
            ]),

            params=NotifyNextcloud.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets) + len(self.groups)
        return max(1, targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Fetch our targets
        results["targets"] = NotifyNextcloud.split_path(results["fullpath"])

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyNextcloud.parse_list(
                results["qsd"]["to"]
            )

        # Allow users to over-ride the Nextcloud version being used
        if "version" in results["qsd"] and len(results["qsd"]["version"]):
            results["version"] = NotifyNextcloud.unquote(
                results["qsd"]["version"]
            )

        # Support URL Prefixes
        if "url_prefix" in results["qsd"] and len(
            results["qsd"]["url_prefix"]
        ):
            results["url_prefix"] = NotifyNextcloud.unquote(
                results["qsd"]["url_prefix"]
            )

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifyNextcloud.unquote(x): NotifyNextcloud.unquote(y)
            for x, y in results["qsd+"].items()
        }

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Create an account https://voip.ms/ if you don't already have one
#
# Enable API and set an API password here:
#   - https://voip.ms/m/api.php
#
# Read more about VoIP.ms API here:
#   - https://voip.ms/m/apidocs.php

import contextlib
from json import loads

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, is_phone_no, parse_phone_no
from .base import NotifyBase


class NotifyVoipms(NotifyBase):
    """A wrapper for VoIPms Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "VoIPms"

    # The services URL
    service_url = "https://voip.ms"

    # The default protocol
    secure_protocol = "voipms"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_voipms"

    # VoIPms uses the http protocol with JSON requests
    notify_url = "https://voip.ms/api/v1/rest.php"

    # The maximum length of the body
    body_maxlen = 160

    # The supported country code by VoIP.ms
    voip_ms_country_code = "1"

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{password}:{email}/{from_phone}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "email": {
                "name": _("User Email"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "from_phone": {
                "name": _("From Phone No"),
                "type": "string",
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "source",
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "alias_of": "from_phone",
            },
        },
    )

    def __init__(self, email, source=None, targets=None, **kwargs):
        """Initialize VoIPms Object."""
        super().__init__(**kwargs)

        # Validate our params here.
        if self.password is None:
            msg = "Password has to be specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # User is the email associated with the account
        result = is_email(email)
        if not result:
            msg = f"An invalid VoIPms user email: ({email}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)
        self.email = result["full_email"]

        # Validate our source Phone #
        result = is_phone_no(source)
        if not result:
            msg = f"An invalid VoIPms source phone # ({source}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Source Phone # only supports +1 country code
        # Allow 7 digit phones (presume they're local with +1 country code)
        if (
            result["country"]
            and result["country"] != self.voip_ms_country_code
        ):
            msg = (
                "VoIPms only supports +1 country code "
                f"({source}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our source phone number (without country code)
        self.source = result["area"] + result["line"]

        # Parse our targets
        self.targets = []

        if targets:
            for target in parse_phone_no(targets):
                # Validate targets and drop bad ones:
                result = is_phone_no(target)

                # Target Phone # only supports +1 country code
                if (
                    result["country"]
                    and result["country"] != self.voip_ms_country_code
                ):
                    self.logger.warning(
                        f"Ignoring invalid phone # ({target}) specified.",
                    )
                    continue

                # store valid phone number
                self.targets.append(result["area"] + result["line"])

        else:
            # Send a message to ourselves
            self.targets.append(self.source)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform VoIPms Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no VoIPms targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        # Prepare our payload
        payload = {
            "api_username": self.email,
            "api_password": self.password,
            "did": self.source,
            "message": body,
            "method": "sendSMS",
            # Gets filled in the loop below
            "dst": None,
        }

        # Create a copy of the targets list
        targets = list(self.targets)

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Add target Phone #
            payload["dst"] = target

            # Some Debug Logging
            self.logger.debug(
                "VoIPms GET URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"VoIPms Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()

            response = {"status": "unknown", "message": ""}

            try:
                r = requests.get(
                    self.notify_url,
                    params=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                with contextlib.suppress(
                        AttributeError, TypeError, ValueError):
                    # Load our JSON object if valid
                    # ValueError = r.content is Unparsable
                    # TypeError = r.content is None
                    # AttributeError = r is None
                    response = loads(r.content)

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyVoipms.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send VoIPms SMS notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                # VoIPms sends 200 OK even if there is an error
                # check if status in response and if it is not success

                if response is not None and response["status"] != "success":
                    self.logger.warning(
                        "Failed to send VoIPms SMS notification to {}: "
                        "status: {}, message: {}".format(
                            target, response["status"], response["message"]
                        )
                    )

                    # Mark our failure
                    has_error = True
                    continue
                else:
                    self.logger.info(
                        f"Sent VoIPms SMS notification to {target}"
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    f"A Connection error occurred sending VoIPms:{target} "
                    "SMS notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.email,
            self.password,
            self.source,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        schemaStr = (
            "{schema}://{password}:{email}/{from_phone}/{targets}/?{params}"
        )
        return schemaStr.format(
            schema=self.secure_protocol,
            email=self.email,
            password=self.pprint(self.password, privacy, safe=""),
            from_phone=self.voip_ms_country_code
            + self.pprint(self.source, privacy, safe=""),
            targets="/".join([
                self.voip_ms_country_code + NotifyVoipms.quote(x, safe="")
                for x in self.targets
            ]),
            params=NotifyVoipms.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        results["targets"] = NotifyVoipms.split_path(results["fullpath"])

        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["source"] = NotifyVoipms.unquote(results["qsd"]["from"])

        elif results["targets"]:
            # The from phone no is the first entry in the list otherwise
            results["source"] = results["targets"].pop(0)

        # Swap user for pass since our input is: password:email
        #   where email is user@hostname (or user@domain)
        user = results["password"]
        password = results["user"]
        results["password"] = password
        results["user"] = user

        results["email"] = "{}@{}".format(
            NotifyVoipms.unquote(user),
            NotifyVoipms.unquote(results["host"]),
        )

        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyVoipms.parse_phone_no(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# 1. Simply visit https://notica.us
# 2. You'll be provided a new variation of the website which will look
#    something like: https://notica.us/?abc123.
#                                         ^
#                                         |
#                                       token
#
#    Your token is actually abc123 (do not include/grab the question mark)
#    You can use that URL as is directly in Apprise, or you can follow
#    the next step which shows you how to assemble the Apprise URL:
#
# 3. With respect to the above, your apprise URL would be:
#       notica://abc123
#
import re

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import validate_regex
from .base import NotifyBase


class NoticaMode:
    """Tracks if we're accessing the notica upstream server or a locally hosted
    one."""

    # We're dealing with a self hosted service
    SELFHOSTED = "selfhosted"

    # We're dealing with the official hosted service at https://notica.us
    OFFICIAL = "official"


# Define our Notica Modes
NOTICA_MODES = (
    NoticaMode.SELFHOSTED,
    NoticaMode.OFFICIAL,
)


class NotifyNotica(NotifyBase):
    """A wrapper for Notica Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Notica"

    # The services URL
    service_url = "https://notica.us/"

    # Insecure protocol (for those self hosted requests)
    protocol = "notica"

    # The default protocol (this is secure for notica)
    secure_protocol = "noticas"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_notica"

    # Notica URL
    notify_url = "https://notica.us/?{token}"

    # Notica does not support a title
    title_maxlen = 0

    # Define object templates
    templates = (
        "{schema}://{token}",
        # Self-hosted notica servers
        "{schema}://{host}/{token}",
        "{schema}://{host}:{port}/{token}",
        "{schema}://{user}@{host}/{token}",
        "{schema}://{user}@{host}:{port}/{token}",
        "{schema}://{user}:{password}@{host}/{token}",
        "{schema}://{user}:{password}@{host}:{port}/{token}",
        # Self-hosted notica servers (with custom path)
        "{schema}://{host}{path}/{token}",
        "{schema}://{host}:{port}/{path}/{token}",
        "{schema}://{user}@{host}/{path}/{token}",
        "{schema}://{user}@{host}:{port}{path}/{token}",
        "{schema}://{user}:{password}@{host}{path}/{token}",
        "{schema}://{user}:{password}@{host}:{port}/{path}/{token}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": r"^\?*(?P<token>[^/]+)\s*$",
            },
            "host": {
                "name": _("Hostname"),
                "type": "string",
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "path": {
                "name": _("Path"),
                "type": "string",
                "map_to": "fullpath",
                "default": "/",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
    }

    def __init__(self, token, headers=None, **kwargs):
        """Initialize Notica Object."""
        super().__init__(**kwargs)

        # Token (associated with project)
        self.token = validate_regex(token)
        if not self.token:
            msg = f"An invalid Notica Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Setup our mode
        self.mode = NoticaMode.SELFHOSTED if self.host else NoticaMode.OFFICIAL

        # prepare our fullpath
        self.fullpath = kwargs.get("fullpath")
        if not isinstance(self.fullpath, str):
            self.fullpath = "/"

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Notica Notification."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
        }

        # Prepare our payload
        payload = f"d:{body}"

        # Auth is used for SELFHOSTED queries
        auth = None

        if self.mode is NoticaMode.OFFICIAL:
            # prepare our notify url
            notify_url = self.notify_url.format(token=self.token)

        else:
            # Prepare our self hosted URL

            # Apply any/all header over-rides defined
            headers.update(self.headers)

            if self.user:
                auth = (self.user, self.password)

            # Set our schema
            schema = "https" if self.secure else "http"

            # Prepare our notify_url
            notify_url = f"{schema}://{self.host}"
            if isinstance(self.port, int):
                notify_url += f":{self.port}"

            notify_url += f"{self.fullpath}?token={self.token}"

        self.logger.debug(
            "Notica POST URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Notica Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url.format(token=self.token),
                data=payload,
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyNotica.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Notica notification:{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Notica notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Notica notification.",
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.mode,
            self.token,
            self.user,
            self.password,
            self.host,
            self.port,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if self.mode == NoticaMode.OFFICIAL:
            # Official URLs are easy to assemble
            return "{schema}://{token}/?{params}".format(
                schema=self.protocol,
                token=self.pprint(self.token, privacy, safe=""),
                params=NotifyNotica.urlencode(params),
            )

        # If we reach here then we are assembling a self hosted URL

        # Append URL parameters from our headers
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Authorization can be used for self-hosted sollutions
        auth = ""

        # Determine Authentication
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyNotica.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyNotica.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return (
            "{schema}://{auth}{hostname}{port}{fullpath}{token}/?{params}"
            .format(
                schema=self.secure_protocol if self.secure else self.protocol,
                auth=auth,
                hostname=NotifyNotica.quote(self.host, safe=""),
                port=(
                    ""
                    if self.port is None or self.port == default_port
                    else f":{self.port}"
                ),
                fullpath=NotifyNotica.quote(self.fullpath, safe="/"),
                token=self.pprint(self.token, privacy, safe=""),
                params=NotifyNotica.urlencode(params),
            )
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get unquoted entries
        entries = NotifyNotica.split_path(results["fullpath"])
        if not entries:
            # If there are no path entries, then we're only dealing with the
            # official website
            results["mode"] = NoticaMode.OFFICIAL

            # Store our token using the host
            results["token"] = NotifyNotica.unquote(results["host"])

            # Unset our host
            results["host"] = None

        else:
            # Otherwise we're running a self hosted instance
            results["mode"] = NoticaMode.SELFHOSTED

            # The last element in the list is our token
            results["token"] = entries.pop()

            # Re-assemble our full path
            results["fullpath"] = (
                "/" if not entries else "/{}/".format("/".join(entries))
            )

            # Add our headers that the user can potentially over-ride if they
            # wish to to our returned result set and tidy entries by unquoting
            # them
            results["headers"] = {
                NotifyNotica.unquote(x): NotifyNotica.unquote(y)
                for x, y in results["qsd+"].items()
            }

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://notica.us/?abc123
        """

        result = re.match(
            r"^https?://notica\.us/?"
            r"\??(?P<token>[^&]+)([&\s]*(?P<params>.+))?$",
            url,
            re.I,
        )

        if result:
            return NotifyNotica.parse_url(
                "{schema}://{token}/{params}".format(
                    schema=NotifyNotica.protocol,
                    token=result.group("token"),
                    params=(
                        ""
                        if not result.group("params")
                        else "?{}".format(result.group("params"))
                    ),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Sources
# - https://dreambox.de/en/
# - https://dream.reichholf.net/wiki/Hauptseite
# - https://dream.reichholf.net/wiki/Enigma2:WebInterface#Message
# - https://github.com/E2OpenPlugins/e2openplugin-OpenWebif
# - https://github.com/E2OpenPlugins/e2openplugin-OpenWebif/wiki/\
#       OpenWebif-API-documentation#message

from json import loads

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from .base import NotifyBase


class Enigma2MessageType:
    # Defines the Enigma2 notification types Apprise can map to
    INFO = 1
    WARNING = 2
    ERROR = 3


# If a mapping fails, the default of Enigma2MessageType.INFO is used
MESSAGE_MAPPING = {
    NotifyType.INFO: Enigma2MessageType.INFO,
    NotifyType.SUCCESS: Enigma2MessageType.INFO,
    NotifyType.WARNING: Enigma2MessageType.WARNING,
    NotifyType.FAILURE: Enigma2MessageType.ERROR,
}


class NotifyEnigma2(NotifyBase):
    """A wrapper for Enigma2 Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Enigma2"

    # The services URL
    service_url = "https://dreambox.de/"

    # The default protocol
    protocol = "enigma2"

    # The default secure protocol
    secure_protocol = "enigma2s"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_enigma2"

    # Enigma2 does not support a title
    title_maxlen = 0

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 1000

    # Throttle a wee-bit to avoid thrashing
    request_rate_per_sec = 0.5

    # Define object templates
    templates = (
        "{schema}://{host}",
        "{schema}://{host}:{port}",
        "{schema}://{user}@{host}",
        "{schema}://{user}@{host}:{port}",
        "{schema}://{user}:{password}@{host}",
        "{schema}://{user}:{password}@{host}:{port}",
        "{schema}://{host}/{fullpath}",
        "{schema}://{host}:{port}/{fullpath}",
        "{schema}://{user}@{host}/{fullpath}",
        "{schema}://{user}@{host}:{port}/{fullpath}",
        "{schema}://{user}:{password}@{host}/{fullpath}",
        "{schema}://{user}:{password}@{host}:{port}/{fullpath}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "fullpath": {
                "name": _("Path"),
                "type": "string",
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "timeout": {
                "name": _("Server Timeout"),
                "type": "int",
                # The number of seconds to display the message for
                "default": 13,
                # -1 means infinit
                "min": -1,
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
    }

    def __init__(self, timeout=None, headers=None, **kwargs):
        """Initialize Enigma2 Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        try:
            self.timeout = int(timeout)
            if self.timeout < self.template_args["timeout"]["min"]:
                # Bulletproof; can't go lower then min value
                self.timeout = self.template_args["timeout"]["min"]

        except (ValueError, TypeError):
            # Use default timeout
            self.timeout = self.template_args["timeout"]["default"]

        self.fullpath = kwargs.get("fullpath")
        if not isinstance(self.fullpath, str):
            self.fullpath = "/"

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol,
            self.user,
            self.password,
            self.host,
            self.port if self.port else (443 if self.secure else 80),
            self.fullpath.rstrip("/"),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "timeout": str(self.timeout),
        }

        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifyEnigma2.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifyEnigma2.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}{fullpath}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            fullpath=NotifyEnigma2.quote(self.fullpath, safe="/"),
            params=NotifyEnigma2.urlencode(params),
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Enigma2 Notification."""

        # prepare Enigma2 Object
        headers = {
            "User-Agent": self.app_id,
        }

        params = {
            "text": body,
            "type": MESSAGE_MAPPING.get(notify_type, Enigma2MessageType.INFO),
            "timeout": self.timeout,
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        # Prepare our message URL
        url += self.fullpath.rstrip("/") + "/api/message"

        self.logger.debug(
            "Enigma2 POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Enigma2 Parameters: {params!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.get(
                url,
                params=params,
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyEnigma2.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Enigma2 notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            # We were able to post our message; now lets evaluate the response
            try:
                # Acquire our result
                result = loads(r.content).get("result", False)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None

                # We could not parse JSON response.
                result = False

            if not result:
                self.logger.warning(
                    "Failed to send Enigma2 notification: "
                    "There was no server acknowledgement."
                )
                self.logger.debug(f"Response Details:\r\n{r.content}")
                # Return; we're done
                return False

            self.logger.info("Sent Enigma2 notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Enigma2 "
                f"notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifyEnigma2.unquote(x): NotifyEnigma2.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # Save timeout value (if specified)
        if "timeout" in results["qsd"] and len(results["qsd"]["timeout"]):
            results["timeout"] = results["qsd"]["timeout"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import hashlib
import hmac
from json import dumps
import re
import time

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase

# Register at https://dingtalk.com
#   - Download their PC based software as it is the only way you can create
#     a custom robot.  You can create a custom robot per group.  You will
#     be provided an access_token that Apprise will need.

# Syntax:
#  dingtalk://{access_token}/
#  dingtalk://{access_token}/{optional_phone_no}
#  dingtalk://{access_token}/{phone_no_1}/{phone_no_2}/{phone_no_N/

# Some Phone Number Detection
IS_PHONE_NO = re.compile(r"^\+?(?P<phone>[0-9\s)(+-]+)\s*$")


class NotifyDingTalk(NotifyBase):
    """A wrapper for DingTalk Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "DingTalk"

    # The services URL
    service_url = "https://www.dingtalk.com/"

    # All notification requests are secure
    secure_protocol = "dingtalk"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_dingtalk"

    # DingTalk API
    notify_url = "https://oapi.dingtalk.com/robot/send?access_token={token}"

    # Do not set title_maxlen as it is set in a property value below
    # since the length varies depending if we are doing a markdown
    # based message or a text based one.
    # title_maxlen = see below @propery defined

    # Define object templates
    templates = (
        "{schema}://{token}/",
        "{schema}://{token}/{targets}/",
        "{schema}://{secret}@{token}/",
        "{schema}://{secret}@{token}/{targets}/",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
            "secret": {
                "name": _("Secret"),
                "type": "string",
                "private": True,
                "regex": (r"^[a-z0-9]+$", "i"),
            },
            "target_phone_no": {
                "name": _("Target Phone No"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "token": {
                "alias_of": "token",
            },
            "secret": {
                "alias_of": "secret",
            },
        },
    )

    def __init__(self, token, targets=None, secret=None, **kwargs):
        """Initialize DingTalk Object."""
        super().__init__(**kwargs)

        # Secret Key (associated with project)
        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"An invalid DingTalk API Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.secret = None
        if secret:
            self.secret = validate_regex(
                secret, *self.template_tokens["secret"]["regex"]
            )
            if not self.secret:
                msg = f"An invalid DingTalk Secret ({token}) was specified."
                self.logger.warning(msg)
                raise TypeError(msg)

        # Parse our targets
        self.targets = []

        for target in parse_list(targets):
            # Validate targets and drop bad ones:
            result = IS_PHONE_NO.match(target)
            if result:
                # Further check our phone # for it's digit count
                result = "".join(re.findall(r"\d+", result.group("phone")))
                if len(result) < 11 or len(result) > 14:
                    self.logger.warning(
                        f"Dropped invalid phone # ({target}) specified.",
                    )
                    continue

                # store valid phone number
                self.targets.append(result)
                continue

            self.logger.warning(
                f"Dropped invalid phone # ({target}) specified.",
            )

        return

    def get_signature(self):
        """Calculates time-based signature so that we can send arbitrary
        messages."""
        timestamp = str(round(time.time() * 1000))
        secret_enc = self.secret.encode("utf-8")
        str_to_sign_enc = f"{timestamp}\n{self.secret}".encode()
        hmac_code = hmac.new(
            secret_enc, str_to_sign_enc, digestmod=hashlib.sha256
        ).digest()
        signature = NotifyDingTalk.quote(base64.b64encode(hmac_code), safe="")
        return timestamp, signature

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform DingTalk Notification."""

        payload = {
            "msgtype": "text",
            "at": {
                "atMobiles": self.targets,
                "isAtAll": False,
            },
        }

        if self.notify_format == NotifyFormat.MARKDOWN:
            payload["markdown"] = {
                "title": title,
                "text": body,
            }

        else:
            payload["text"] = {
                "content": body,
            }

        # Our Notification URL
        notify_url = self.notify_url.format(token=self.token)

        params = None
        if self.secret:
            timestamp, signature = self.get_signature()
            params = {
                "timestamp": timestamp,
                "sign": signature,
            }

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Some Debug Logging
        self.logger.debug(
            "DingTalk URL:"
            f" {notify_url} (cert_verify={self.verify_certificate})"
        )
        self.logger.debug(f"DingTalk Payload: {payload}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=dumps(payload),
                headers=headers,
                params=params,
                verify=self.verify_certificate,
            )

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyDingTalk.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send DingTalk notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")
                return False

            else:
                self.logger.info("Sent DingTalk notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occured sending DingTalk notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        return True

    @property
    def title_maxlen(self):
        """The title isn't used when not in markdown mode."""
        return (
            NotifyBase.title_maxlen
            if self.notify_format == NotifyFormat.MARKDOWN
            else 0
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any arguments set
        args = {
            "format": self.notify_format,
            "overflow": self.overflow_mode,
            "verify": "yes" if self.verify_certificate else "no",
        }

        return "{schema}://{secret}{token}/{targets}/?{args}".format(
            schema=self.secure_protocol,
            secret=(
                ""
                if not self.secret
                else "{}@".format(
                    self.pprint(
                        self.secret, privacy, mode=PrivacyMode.Secret, safe=""
                    )
                )
            ),
            token=self.pprint(self.token, privacy, safe=""),
            targets="/".join(
                [NotifyDingTalk.quote(x, safe="") for x in self.targets]
            ),
            args=NotifyDingTalk.urlencode(args),
        )

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.secret, self.token)

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.targets)
        return targets if targets > 0 else 1

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to
        substantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        results["token"] = NotifyDingTalk.unquote(results["host"])

        # if a user has been defined, use it's value as the secret
        if results.get("user"):
            results["secret"] = results.get("user")

        # Get our entries; split_path() looks after unquoting content for us
        # by default
        results["targets"] = NotifyDingTalk.split_path(results["fullpath"])

        # Support the use of the `token` keyword argument
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifyDingTalk.unquote(results["qsd"]["token"])

        # Support the use of the `secret` keyword argument
        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            results["secret"] = NotifyDingTalk.unquote(
                results["qsd"]["secret"]
            )

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyDingTalk.parse_list(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from html.parser import HTMLParser
import re

from markdown import markdown

from .common import NotifyFormat
from .url import URLBase


def convert_between(from_format, to_format, content):
    """Converts between different suported formats. If no conversion exists, or
    the selected one fails, the original text will be returned.

    This function returns the content translated (if required)
    """

    converters = {
        (NotifyFormat.MARKDOWN, NotifyFormat.HTML): markdown_to_html,
        (NotifyFormat.TEXT, NotifyFormat.HTML): text_to_html,
        (NotifyFormat.HTML, NotifyFormat.TEXT): html_to_text,
        # For now; use same converter for Markdown support
        (NotifyFormat.HTML, NotifyFormat.MARKDOWN): html_to_text,
    }

    convert = converters.get((from_format, to_format))
    return convert(content) if convert else content


def markdown_to_html(content):
    """Converts specified content from markdown to HTML."""
    return markdown(
        content,
        extensions=["markdown.extensions.nl2br", "markdown.extensions.tables"],
    )


def text_to_html(content):
    """Converts specified content from plain text to HTML."""

    # First eliminate any carriage returns
    return URLBase.escape_html(content, convert_new_lines=True)


def html_to_text(content):
    """Converts a content from HTML to plain text."""

    parser = HTMLConverter()
    parser.feed(content)
    parser.close()
    return parser.converted


class HTMLConverter(HTMLParser):
    """An HTML to plain text converter tuned for email messages."""

    # The following tags must start on a new line
    BLOCK_TAGS = (
        "p",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "div",
        "td",
        "th",
        "code",
        "pre",
        "label",
        "li",
    )

    # the folowing tags ignore any internal text
    IGNORE_TAGS = (
        "form",
        "input",
        "textarea",
        "select",
        "ul",
        "ol",
        "style",
        "link",
        "meta",
        "title",
        "html",
        "head",
        "script",
    )

    # Condense Whitespace
    WS_TRIM = re.compile(r"[\s]+", re.DOTALL | re.MULTILINE)

    # Sentinel value for block tag boundaries, which may be consolidated into a
    # single line break.
    BLOCK_END = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Shoudl we store the text content or not?
        self._do_store = True

        # Initialize internal result list
        self._result = []

        # Initialize public result field (not populated until close() is
        # called)
        self.converted = ""

    def close(self):
        string = "".join(self._finalize(self._result))
        self.converted = string.strip()

    def _finalize(self, result):
        """Combines and strips consecutive strings, then converts consecutive
        block ends into singleton newlines.

        [ {be} " Hello " {be} {be} " World!" ] -> "\nHello\nWorld!"
        """

        # None means the last visited item was a block end.
        accum = None

        for item in result:
            if item == self.BLOCK_END:
                # Multiple consecutive block ends; do nothing.
                if accum is None:
                    continue

                # First block end; yield the current string, plus a newline.
                yield accum.strip() + "\n"
                accum = None

            # Multiple consecutive strings; combine them.
            elif accum is not None:
                accum += item

            # First consecutive string; store it.
            else:
                accum = item

        # Yield the last string if we have not already done so.
        if accum is not None:
            yield accum.strip()

    def handle_data(self, data, *args, **kwargs):
        """Store our data if it is not on the ignore list."""

        # initialize our previous flag
        if self._do_store:

            # Tidy our whitespace
            content = self.WS_TRIM.sub(" ", data)
            self._result.append(content)

    def handle_starttag(self, tag, attrs):
        """Process our starting HTML Tag."""
        # Toggle initial states
        self._do_store = tag not in self.IGNORE_TAGS

        if tag in self.BLOCK_TAGS:
            self._result.append(self.BLOCK_END)

        if tag == "li":
            self._result.append("- ")

        elif tag == "br":
            self._result.append("\n")

        elif tag == "hr":
            if self._result and isinstance(self._result[-1], str):
                self._result[-1] = self._result[-1].rstrip(" ")
            else:
                pass

            self._result.append("\n---\n")

        elif tag == "blockquote":
            self._result.append(" >")

    def handle_endtag(self, tag):
        """Edge case handling of open/close tags."""
        self._do_store = True

        if tag in self.BLOCK_TAGS:
            self._result.append(self.BLOCK_END)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from .base import NotifyBase

# For API Details see:
# https://kb.synology.com/en-au/DSM/help/Chat/chat_integration


class NotifySynology(NotifyBase):
    """A wrapper for Synology Chat Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Synology Chat"

    # The services URL
    service_url = "https://www.synology.com/"

    # The default protocol
    protocol = "synology"

    # The default secure protocol
    secure_protocol = "synologys"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_synology_chat"

    # Title is to be part of body
    title_maxlen = 0

    # Disable throttle rate for Synology requests since they are normally
    # local anyway
    request_rate_per_sec = 0

    # Define object templates
    templates = (
        "{schema}://{host}/{token}",
        "{schema}://{host}:{port}/{token}",
        "{schema}://{user}@{host}/{token}",
        "{schema}://{user}@{host}:{port}/{token}",
        "{schema}://{user}:{password}@{host}/{token}",
        "{schema}://{user}:{password}@{host}:{port}/{token}",
    )

    # Define our tokens; these are the minimum tokens required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "token": {
                "name": _("Token"),
                "type": "string",
                "required": True,
                "private": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "file_url": {
                "name": _("Upload"),
                "type": "string",
            },
            "token": {
                "alias_of": "token",
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "headers": {
            "name": _("HTTP Header"),
            "prefix": "+",
        },
    }

    def __init__(self, token=None, headers=None, file_url=None, **kwargs):
        """Initialize Synology Chat Object.

        headers can be a dictionary of key/value pairs that you want to
        additionally include as part of the server headers to post with
        """
        super().__init__(**kwargs)

        self.token = token
        if not self.token:
            msg = f"An invalid Synology Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.fullpath = kwargs.get("fullpath")

        # A URL to an attachment you want to upload (must be less then 32MB
        # Acording to API details (at the time of writing plugin)
        self.file_url = file_url

        self.headers = {}
        if headers:
            # Store our extra headers
            self.headers.update(headers)

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.host,
            self.port,
            self.token,
            self.fullpath.rstrip("/"),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {}

        if self.file_url:
            params["file_url"] = self.file_url

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Append our headers into our parameters
        params.update({f"+{k}": v for k, v in self.headers.items()})

        # Determine Authentication
        auth = ""
        if self.user and self.password:
            auth = "{user}:{password}@".format(
                user=NotifySynology.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )
        elif self.user:
            auth = "{user}@".format(
                user=NotifySynology.quote(self.user, safe=""),
            )

        default_port = 443 if self.secure else 80

        return (
            "{schema}://{auth}{hostname}{port}/{token}"
            "{fullpath}?{params}".format(
                schema=self.secure_protocol if self.secure else self.protocol,
                auth=auth,
                # never encode hostname since we're expecting it to be a valid
                # one
                hostname=self.host,
                port=(
                    ""
                    if self.port is None or self.port == default_port
                    else f":{self.port}"
                ),
                token=self.pprint(self.token, privacy, safe=""),
                fullpath=(
                    NotifySynology.quote(self.fullpath, safe="/")
                    if self.fullpath
                    else "/"
                ),
                params=NotifySynology.urlencode(params),
            )
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Synology Chat Notification."""

        # Prepare HTTP Headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "*/*",
        }

        # Apply any/all header over-rides defined
        headers.update(self.headers)

        # prepare Synology Object
        payload = {
            "text": body,
        }

        if self.file_url:
            payload["file_url"] = self.file_url

        # Prepare our parameters
        params = {
            "api": "SYNO.Chat.External",
            "method": "incoming",
            "version": 2,
            "token": self.token,
        }

        auth = None
        if self.user:
            auth = (self.user, self.password)

        # Set our schema
        schema = "https" if self.secure else "http"

        url = f"{schema}://{self.host}"
        if isinstance(self.port, int):
            url += f":{self.port}"

        # Prepare our Synology API URL
        url += self.fullpath + "/webapi/entry.cgi"

        self.logger.debug(
            "Synology Chat POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Synology Chat Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                url,
                data=f"payload={dumps(payload)}",
                params=params,
                headers=headers,
                auth=auth,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code < 200 or r.status_code >= 300:
                # We had a problem
                status_str = NotifySynology.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Synology Chat %s notification: "
                    "%serror=%s.",
                    status_str,
                    ", " if status_str else "",
                    str(r.status_code),
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Synology Chat notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Synology "
                f"Chat notification to {self.host}."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Add our headers that the user can potentially over-ride if they wish
        # to to our returned result set and tidy entries by unquoting them
        results["headers"] = {
            NotifySynology.unquote(x): NotifySynology.unquote(y)
            for x, y in results["qsd+"].items()
        }

        # Set our token if found as an argument
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifySynology.unquote(results["qsd"]["token"])

        else:
            # Get unquoted entries
            entries = NotifySynology.split_path(results["fullpath"])
            if entries:
                # Pop the first element
                results["token"] = entries.pop(0)

                # Update our fullpath to not include our token
                results["fullpath"] = results["fullpath"][
                    len(results["token"]) + 1:
                ]

        # Set upload/file_url if not otherwise set
        if "file_url" in results["qsd"] and len(results["qsd"]["file_url"]):
            results["file_url"] = NotifySynology.unquote(
                results["qsd"]["file_url"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from itertools import chain
from json import dumps, loads
import re

import requests

from ..common import NotifyFormat, NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, parse_list
from .base import NotifyBase

IS_CHANNEL = re.compile(r"^#(?P<name>[A-Za-z0-9_-]+)$")
IS_USER = re.compile(r"^@(?P<name>[A-Za-z0-9._-]+)$")
IS_ROOM_ID = re.compile(r"^(?P<name>[A-Za-z0-9]+)$")

# Extend HTTP Error Messages
RC_HTTP_ERROR_MAP = {
    400: "Channel/RoomId is wrong format, or missing from server.",
    401: "Authentication tokens provided is invalid or missing.",
}


class RocketChatAuthMode:
    """The Chat Authentication mode is detected."""

    # providing a webhook
    WEBHOOK = "webhook"

    # Support token submission
    TOKEN = "token"

    # Providing a username and password (default)
    BASIC = "basic"


# Define our authentication modes
ROCKETCHAT_AUTH_MODES = (
    RocketChatAuthMode.WEBHOOK,
    RocketChatAuthMode.TOKEN,
    RocketChatAuthMode.BASIC,
)


class NotifyRocketChat(NotifyBase):
    """A wrapper for Notify Rocket.Chat Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Rocket.Chat"

    # The services URL
    service_url = "https://rocket.chat/"

    # The default protocol
    protocol = "rocket"

    # The default secure protocol
    secure_protocol = "rockets"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_rocketchat"

    # Allows the user to specify the NotifyImageSize object; this is supported
    # through the webhook
    image_size = NotifyImageSize.XY_128

    # The title is not used
    title_maxlen = 0

    # The maximum size of the message
    body_maxlen = 1000

    # Default to markdown
    notify_format = NotifyFormat.MARKDOWN

    # Define object templates
    templates = (
        "{schema}://{user}:{password}@{host}:{port}/{targets}",
        "{schema}://{user}:{password}@{host}/{targets}",
        "{schema}://{user}:{token}@{host}:{port}/{targets}",
        "{schema}://{user}:{token}@{host}/{targets}",
        "{schema}://{webhook}@{host}",
        "{schema}://{webhook}@{host}:{port}",
        "{schema}://{webhook}@{host}/{targets}",
        "{schema}://{webhook}@{host}:{port}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("Username"),
                "type": "string",
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
            },
            "token": {
                "name": _("API Token"),
                "map_to": "password",
                "private": True,
            },
            "webhook": {
                "name": _("Webhook"),
                "type": "string",
            },
            "target_channel": {
                "name": _("Target Channel"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "prefix": "@",
                "map_to": "targets",
            },
            "target_room": {
                "name": _("Target Room ID"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "mode": {
                "name": _("Webhook Mode"),
                "type": "choice:string",
                "values": ROCKETCHAT_AUTH_MODES,
            },
            "avatar": {
                "name": _("Use Avatar"),
                "type": "bool",
                "default": False,
            },
            "webhook": {
                "alias_of": "webhook",
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(
        self, webhook=None, targets=None, mode=None, avatar=None, **kwargs
    ):
        """Initialize Notify Rocket.Chat Object."""
        super().__init__(**kwargs)

        # Set our schema
        self.schema = "https" if self.secure else "http"

        # Prepare our URL
        self.api_url = f"{self.schema}://{self.host}"

        if isinstance(self.port, int):
            self.api_url += f":{self.port}"

        # Initialize channels list
        self.channels = []

        # Initialize room list
        self.rooms = []

        # Initialize user list (webhook only)
        self.users = []

        # Assign our webhook (if defined)
        self.webhook = webhook

        # Used to track token headers upon authentication (if successful)
        # This is only used if not on webhook mode
        self.headers = {}

        # Authentication mode
        self.mode = None if not isinstance(mode, str) else mode.lower()

        if self.mode and self.mode not in ROCKETCHAT_AUTH_MODES:
            msg = f"The authentication mode specified ({mode}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Detect our mode if it wasn't specified
        if not self.mode:
            if self.webhook is not None:
                # Just a username was specified, we treat this as a webhook
                self.mode = RocketChatAuthMode.WEBHOOK
            elif self.password and len(self.password) > 32:
                self.mode = RocketChatAuthMode.TOKEN
            else:
                self.mode = RocketChatAuthMode.BASIC

            self.logger.debug(
                "Auto-Detected Rocketchat Auth Mode: %s", self.mode
            )

        if self.mode in (
            RocketChatAuthMode.BASIC,
            RocketChatAuthMode.TOKEN,
        ) and not (self.user and self.password):
            # Username & Password is required for Rocket Chat to work
            msg = "No Rocket.Chat {} was specified.".format(
                "user/pass combo"
                if self.mode == RocketChatAuthMode.BASIC
                else "user/apikey"
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        elif self.mode == RocketChatAuthMode.WEBHOOK and not self.webhook:
            msg = "No Rocket.Chat Incoming Webhook was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        if self.mode == RocketChatAuthMode.TOKEN:
            # Set our headers for further communication
            self.headers.update({
                "X-User-Id": self.user,
                "X-Auth-Token": self.password,
            })

        # Validate recipients and drop bad ones:
        for recipient in parse_list(targets):
            result = IS_CHANNEL.match(recipient)
            if result:
                # store valid device
                self.channels.append(result.group("name"))
                continue

            result = IS_ROOM_ID.match(recipient)
            if result:
                # store valid room
                self.rooms.append(result.group("name"))
                continue

            result = IS_USER.match(recipient)
            if result:
                # store valid room
                self.users.append(result.group("name"))
                continue

            self.logger.warning(
                f"Dropped invalid channel/room/user ({recipient}) specified.",
            )

        if (
            self.mode == RocketChatAuthMode.BASIC
            and len(self.rooms) == 0
            and len(self.channels) == 0
        ):
            msg = "No Rocket.Chat room and/or channels specified to notify."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Prepare our avatar setting
        # - if specified; that trumps all
        # - if not specified and we're dealing with a basic setup, the Avatar
        #   is disabled by default. This is because if the account doesn't
        #   have the bot flag set on it it won't work as documented here:
        #       https://developer.rocket.chat/api/rest-api/endpoints\
        #             /team-collaboration-endpoints/chat/postmessage
        # - Otherwise if we're a webhook, we enable the avatar by default
        #   (if not otherwise specified) since it will work nicely.
        # Place an avatar image to associate with our content
        if self.mode == RocketChatAuthMode.BASIC:
            self.avatar = False if avatar is None else avatar

        else:  # self.mode == RocketChatAuthMode.WEBHOOK:
            self.avatar = True if avatar is None else avatar

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.host,
            self.port if self.port else (443 if self.secure else 80),
            self.user,
            (
                self.password
                if self.mode
                in (RocketChatAuthMode.BASIC, RocketChatAuthMode.TOKEN)
                else self.webhook
            ),
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "avatar": "yes" if self.avatar else "no",
            "mode": self.mode,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Determine Authentication
        if self.mode in (RocketChatAuthMode.BASIC, RocketChatAuthMode.TOKEN):
            auth = "{user}:{password}@".format(
                user=NotifyRocketChat.quote(self.user, safe=""),
                password=self.pprint(
                    self.password, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )

        else:
            auth = "{user}{webhook}@".format(
                user=(
                    "{}:".format(NotifyRocketChat.quote(self.user, safe=""))
                    if self.user
                    else ""
                ),
                webhook=self.pprint(
                    self.webhook, privacy, mode=PrivacyMode.Secret, safe=""
                ),
            )

        default_port = 443 if self.secure else 80
        return "{schema}://{auth}{hostname}{port}/{targets}/?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            auth=auth,
            # never encode hostname since we're expecting it to be a valid one
            hostname=self.host,
            port=(
                ""
                if self.port is None or self.port == default_port
                else f":{self.port}"
            ),
            targets="/".join([
                NotifyRocketChat.quote(x, safe="@#")
                for x in chain(
                    # Channels are prefixed with a pound/hashtag symbol
                    [f"#{x}" for x in self.channels],
                    # Rooms are as is
                    self.rooms,
                    # Users
                    [f"@{x}" for x in self.users],
                )
            ]),
            params=NotifyRocketChat.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        targets = len(self.channels) + len(self.rooms) + len(self.users)
        return targets if targets > 0 else 1

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Wrapper to _send since we can alert more then one channel."""

        # Call the _send_ function applicable to whatever mode we're in
        # - calls _send_webhook_notification if the mode variable is set
        # - calls _send_basic_notification if the mode variable is not set
        return getattr(
            self,
            "_send_{}_notification".format(
                RocketChatAuthMode.WEBHOOK
                if self.mode == RocketChatAuthMode.WEBHOOK
                else RocketChatAuthMode.BASIC
            ),
        )(body=body, title=title, notify_type=notify_type, **kwargs)

    def _send_webhook_notification(
        self, body, title="", notify_type=NotifyType.INFO, **kwargs
    ):
        """Sends a webhook notification."""

        # Our payload object
        payload = self._payload(body, title, notify_type)

        # Assemble our webhook URL
        path = f"hooks/{self.webhook}"

        # Build our list of channels/rooms/users (if any identified)
        targets = [f"@{u}" for u in self.users]
        targets.extend([f"#{c}" for c in self.channels])
        targets.extend([f"{r}" for r in self.rooms])

        if len(targets) == 0:
            # We can take an early exit
            return self._send(
                payload, notify_type=notify_type, path=path, **kwargs
            )

        # Otherwise we want to iterate over each of the targets

        # Initiaize our error tracking
        has_error = False

        while len(targets):
            # Retrieve our target
            target = targets.pop(0)

            # Assign our channel/room/user
            payload["channel"] = target

            if not self._send(
                payload, notify_type=notify_type, path=path, **kwargs
            ):

                # toggle flag
                has_error = True

        return not has_error

    def _send_basic_notification(
        self, body, title="", notify_type=NotifyType.INFO, **kwargs
    ):
        """Authenticates with the server using a user/pass combo for
        notifications."""
        # Track whether we authenticated okay

        if self.mode == RocketChatAuthMode.BASIC and not self.login():
            return False

        # prepare JSON Object
        _payload = self._payload(body, title, notify_type)

        # Initiaize our error tracking
        has_error = False

        # Build our list of channels/rooms/users (if any identified)
        channels = [f"@{u}" for u in self.users]
        channels.extend([f"#{c}" for c in self.channels])

        # Create a copy of our channels to notify against
        payload = _payload.copy()
        while len(channels) > 0:
            # Get Channel
            channel = channels.pop(0)
            payload["channel"] = channel

            if not self._send(payload, notify_type=notify_type, **kwargs):

                # toggle flag
                has_error = True

        # Create a copy of our room id's to notify against
        rooms = list(self.rooms)
        payload = _payload.copy()
        while len(rooms):
            # Get Room
            room = rooms.pop(0)
            payload["roomId"] = room

            if not self._send(payload, notify_type=notify_type, **kwargs):

                # toggle flag
                has_error = True

        if self.mode == RocketChatAuthMode.BASIC:
            # logout
            self.logout()

        return not has_error

    def _payload(self, body, title="", notify_type=NotifyType.INFO):
        """Prepares a payload object."""
        # prepare JSON Object
        payload = {
            "text": body,
        }

        # apply our images if they're set to be displayed
        image_url = self.image_url(notify_type)
        if self.avatar and image_url:
            payload["avatar"] = image_url

        return payload

    def _send(
        self, payload, notify_type, path="api/v1/chat.postMessage", **kwargs
    ):
        """Perform Notify Rocket.Chat Notification."""

        api_url = f"{self.api_url}/{path}"

        self.logger.debug(
            "Rocket.Chat POST URL:"
            f" {api_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Rocket.Chat Payload: {payload!s}")

        # Copy our existing headers
        headers = self.headers.copy()

        # Apply minimum headers
        headers.update({
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        })

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                api_url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyRocketChat.http_response_code_lookup(
                    r.status_code, RC_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Rocket.Chat {}:notification: "
                    "{}{}error={}.".format(
                        self.mode,
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info(f"Sent Rocket.Chat {self.mode}:notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Rocket.Chat "
                f"{self.mode}:notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    def login(self):
        """Login to our server."""

        payload = {
            "username": self.user,
            "password": self.password,
        }

        api_url = "{}/{}".format(self.api_url, "api/v1/login")

        try:
            r = requests.post(
                api_url,
                data=payload,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyRocketChat.http_response_code_lookup(
                    r.status_code, RC_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to authenticate {} with Rocket.Chat: "
                    "{}{}error={}.".format(
                        self.user,
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.debug("Rocket.Chat authentication successful")
                response = loads(r.content)
                if response.get("status") != "success":
                    self.logger.warning(
                        f"Could not authenticate {self.user} with Rocket.Chat."
                    )
                    return False

                # Set our headers for further communication
                self.headers["X-Auth-Token"] = response.get(
                    "data", {"authToken": None}
                ).get("authToken")
                self.headers["X-User-Id"] = response.get(
                    "data", {"userId": None}
                ).get("userId")

        except (AttributeError, TypeError, ValueError):
            # Our response was not the JSON type we had expected it to be
            # - ValueError = r.content is Unparsable
            # - TypeError = r.content is None
            # - AttributeError = r is None
            self.logger.warning(
                f"A commuication error occurred authenticating {self.user} on "
                "Rocket.Chat."
            )
            return False

        except requests.RequestException as e:
            self.logger.warning(
                f"A connection error occurred authenticating {self.user} on "
                "Rocket.Chat."
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        return True

    def logout(self):
        """Logout of our server."""

        api_url = "{}/{}".format(self.api_url, "api/v1/logout")

        try:
            r = requests.post(
                api_url,
                headers=self.headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyRocketChat.http_response_code_lookup(
                    r.status_code, RC_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to logoff {} from Rocket.Chat: "
                    "{}{}error={}.".format(
                        self.user,
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.debug(
                    f"Rocket.Chat log off successful; response {r.content}."
                )

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred logging off the "
                "Rocket.Chat server"
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        try:
            # Attempt to detect the webhook (if specified in the URL)
            # If no webhook is specified, then we just pass along as if nothing
            # happened. However if we do find a webhook, we want to rebuild our
            # URL without it since it conflicts with standard URLs. Support
            # %2F since that is a forward slash escaped

            # rocket://webhook@host
            # rocket://user:webhook@host
            match = re.match(
                r"^\s*(?P<schema>[^:]+://)((?P<user>[^:]+):)?"
                r"(?P<webhook>[a-z0-9]+(/|%2F)"
                r"[a-z0-9]+)\@(?P<url>.+)$",
                url,
                re.I,
            )

        except TypeError:
            # Not a string
            return None

        if match:
            # Re-assemble our URL without the webhook
            url = "{schema}{user}{url}".format(
                schema=match.group("schema"),
                user=(
                    "{}@".format(match.group("user"))
                    if match.group("user")
                    else ""
                ),
                url=match.group("url"),
            )

        results = NotifyBase.parse_url(url)
        if not results:
            # We're done early as we couldn't load the results
            return results

        if match:
            # store our webhook
            results["webhook"] = NotifyRocketChat.unquote(
                match.group("webhook")
            )

            # Take on the password too in the event we're in basic mode
            # We do not unquote() as this is done at a later state
            results["password"] = match.group("webhook")

        # Apply our targets
        results["targets"] = NotifyRocketChat.split_path(results["fullpath"])

        # The user may have forced the mode
        if "mode" in results["qsd"] and len(results["qsd"]["mode"]):
            results["mode"] = NotifyRocketChat.unquote(results["qsd"]["mode"])

        # avatar icon
        if "avatar" in results["qsd"] and len(results["qsd"]["avatar"]):
            results["avatar"] = parse_bool(results["qsd"].get("avatar", True))

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyRocketChat.parse_list(
                results["qsd"]["to"]
            )

        # The 'webhook' over-ride (if specified)
        if "webhook" in results["qsd"] and len(results["qsd"]["webhook"]):
            results["webhook"] = NotifyRocketChat.unquote(
                results["qsd"]["webhook"]
            )

        return results
